!************************************************************
! 
!
!***********************************************************
module omphotom_f1 
use dal
use cal
use sasdal
use oal
use errorHandling
use frametime_mod, only : getFrameTime ! mssllib
use getFilterID_mod
use omutility_mod      ! In mssllib various fits access routines
use omconstants_mod    ! In mssllib defines various constants
use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
use omparameters_mod 
use omphotometry_mod   ! In mssllib for doing aperture photometry
use ommagnitude_mod    ! In mssllib for computing instrumental magnitude 
use ominterpolate_mod
use omnumerics_mod
use ompsf_mod
use leastsquares_mod
use omminim_mod
use psfphotometry_mod
implicit none
 
!********************************************************************
! Global variables
!********************************************************************



    real(single), private :: myApertureRadius = 0.0, myCalRadius = 0.0

    real(single), private, pointer :: myImage(:, :) => null()

    real(single), private, pointer :: myBackgroundImage(:, :) => null()

    integer(int16), private, pointer :: mylevelImage(:, :) => null()

    logical(boolean), private, pointer :: myQualityImage(:, :) => null()

 
    real(single), private :: exposure = 0.0
    real(double), private :: plateScale 
    real(double), private :: mjd = 0.0
    character(10), private :: filter     
   
    real(single), private :: innerBackgroundRadius = 0.0, outerBackgroundRadius = 0.0

    real(single), private :: meanBackground

    type(SourceParameters), private, pointer    :: sources(:) => null()
    type(SourceParameters), private, pointer    :: oldSources(:) => null()

    logical, private  :: myScaleCounts = .false.
    real(double), private ::  myScaleCountsRatio = 1.0 
   
    integer(int32), private :: nSources = 0

    logical :: useBackgroundImage = .false. 

    !***************************************************************************
    ! The following is for FAST data, if using a separate image for background
    !***************************************************************************
    integer :: x0Image, y0Image, imageBinx, imageBiny 

    integer :: x0Background, y0Background, backgroundBinx, backgroundBinY


interface extendedSourcePhotometry
    module procedure extendedSourcePhotometry1, extendedSourcePhotometry2 
end interface extendedSourcePhotometry


contains



!*********************************************************************
! Info: Initialises the fortran part of omsource.
!       It loads the image pdata and properties, and -if any- it also
!       creates a source list from an already existing sources dataset
!*********************************************************************
subroutine initialise_data( n )
    integer, intent(out) :: n
    real(single) :: sigma 
       
    type(dataSetT) :: set, backgroundset
    logical :: binBpe = .false.
   

    ! --- extract the image from the file
    set = dataSet(stringParameter("imageset"), read)

     
    exposure = real32Attribute( set, "EXPOSURE")  
    filter = stringAttribute( set, "FILTER" )   

    ! compute binning
    binBpe = booleanAttribute(set, "BINBPE")
    binx     = 2 ** int32Attribute(set, "BINAX1")
    biny     = 2 ** int32Attribute(set, "BINAX2")


    if (binBpe) then
        binx= binx * 2
        biny= biny * 2
    end if
    


    ! CCD frame-time and dead-fraction
   
          
    frameTime = real64Attribute(set, "FRAMTIME") / 1000.0
    deadFrac = real64Attribute(set, "DEADFRAC")


    !*************************************************************************
    ! Need the modified Julian date for time-dependent sensitivity correction
    !************************************************************************
     mjd = ModifiedJulianDate( trim( stringAttribute(set, "DATE-OBS")))


    call CAL_setState(OM, filterId = filterIdToInt(trim( filter ) ))

    call setSignificanceRadius( filter ) 
      

      ! now init radii once, and once only, and bin them now
    myApertureRadius = 12.0                 ! NEED THIS IN UNBINND PIXELS

    myCalRadius = CAL_omGetApertureRadius() ! NEED THIS IN UNBINNED PIXELS


    if (myCalRadius .gt. myApertureRadius) then
        !**********************************************
        ! Remember that counts have to be extrapolated.
        !**********************************************
        myScaleCounts = .true.
        myScaleCountsRatio = (myCalRadius / myApertureRadius) ** 2 

    else
        myScaleCounts = .false. 




    end if



    !************************************* 
    ! Get the Plate scale (arcsecs/pixel)
    !*************************************
    call CAL_omGetPlateScale(0.55_double, plateScale)

    ! Get the pixels
    call createImageArray( set, myImage )



    ! Get the quality of the pixels
!    call createQualityArray( set, myQualityImage )
   
  if(associated(myQualityImage)) then
        deallocate(myQualityImage)
        nullify(myQualityIMage)
    endif
   

    nx = size(myImage, 1)
    ny = size(myImage, 2)
 



    myQualityImage => getQualityArray(set, nx, ny)



    call release( set )  


    ! Compute the background mode

 
 
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1 
    
    meanBackground = backgroundMode(myImage, int(nx, int16), int(ny, int16), 2.0, sigma) 


    myLevelImage => allocate2dInt16(0, nx - 1, 0, ny - 1, "myLevelImage")


     
    call loadSources(sources, stringParameter("oldsrclistset"))
    n = nSources




    !***********************************************************
    !
    !**********************************************************
    call refineSourceNumImage(sources, int(nsources, int16), mylevelimage)


    !********************************************************************************
    ! Check to see if there is a background image specified and see if it not the
    ! actual image
    !********************************************************************************
    
    if( len( trim(stringParameter("backgroundimageset") )) > 2 .and. &
             trim(stringParameter("backgroundimageset")) .ne.  trim(stringParameter("imageset"))) then 

        useBackgroundImage = .true.
        !extract the image from the file
        backgroundSet = dataSet(stringParameter("backgroundimageset"), read)
        call createImageArray( backgroundset, myBackgroundImage )
        call release(backgroundSet)  


        if( size(myBackgroundImage, 1) .ne. nx) then

            !***************************************************************
            ! Need to make the image the same size as the background image
            !***************************************************************

            call getParameters
            
            fastBackground = .true.  

        else

            fastBackground = .false.  

        end if

    else
        !***************************************************************
        ! The background image is the same as the image
        !***************************************************************

        useBackgroundImage = .false.
        myBackgroundImage => myImage
        fastBackground = .false.  

    end if



    !*****************************************************************
    !
    !*****************************************************************

    call initialisePSFPhotometry(myimage, filter, myapertureRadius, mycalRadius)

    


!call SaveImage(mypsfimage, "psfimage.fits", "test")


end subroutine initialise_data

!*********************************************************************************
! subroutine loadSources
! Info: loads sources found in dataset(setName), and returns a newly created vector
!       of sources.
! 
!********************************************************************************
subroutine loadSources(sources, setName)

    type(SourceParameters), pointer    :: sources(:)
    character(len=*), intent(in) :: setName
   

  
    type(DataSetT)        ::  set
    type(TableT)          :: inTableH
    
      
    integer i
      ! --- don't bother if none available
    if (trim(setName) == "" ) return

    ! Read in the OM OSW Source List file.
    set = dataSet(setName, read)
    exposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  

    if (.not. hasBlock(set, "SRCLIST")) then
     

        ! Incompatible file - can't continue
        write (errstr, *) "Could not find any compatibile blocks in set ", trim(setName)
        call fatal("BadFileType")

    else
        ! Get a handle on the source list binary table extension.
        inTableH = table(set, "SRCLIST")

    end if


    ! Get the number of rows in the source list binary table extension.
    nSources = numberOfRows(inTableH)

    if (nSources == 0) then
       
        call release(inTableH)
        call release(set)
        return

    end if



    innerBackgroundRadius =  GetReal32Attribute(set, "INNBCKAN")
    outerBackgroundRadius =  GetReal32Attribute(set, "OUTBCKAN")



    allocate(sources(1 : nSources) )
 
   


    ! Get the data for each source from the source list file
  
      
    call LoadReal32Column(inTableH, "XPOS", sources%xpos)
  
    call LoadReal32Column(inTableH, "YPOS", sources%ypos)
       
  
    call LoadReal32Column(inTableH, "RATE", sources%rawCountRate)
  
  
    call LoadReal32Column(inTableH, "RATE_ERR", sources%rawCountRateErr)
  

    call LoadReal32Column(inTableH, "CORR_RATE", sources%corrCountRate)
  
  
    call LoadReal32Column(inTableH, "CORR_RATE_ERR", sources%corrCountRateErr)
   

    call LoadReal32Column(inTableH, "BK_COILOSS", sources%semibErr)

    call LoadReal32Column(inTableH, "SPB_COILOSS", sources%semiaErr)

    call LoadReal32Column(inTableH, "PSF1_CORR", sources%psf1Correction)
  
  
    call LoadReal32Column(inTableH, "PSF2_CORR", sources%psf2Correction)
  


    call LoadReal32Column(inTableH, "BACKGROUND_RATE",  sources%rawBackgroundCountRate)
    
  
    call LoadReal32Column(inTableH, "CORR_BACKGROUND_RATE",  sources%corrbackgroundCountRate)
   

    call LoadReal32Column(inTableH, "MAG", sources%magnitude)
    
  
    call LoadReal32Column(inTableH, "MAGERR", sources%magnitudeErr)
   
  
    call LoadReal32Column(inTableH, "SIGNIFICANCE", sources%significance)
   
  
    call LoadReal32Column(inTableH, "APERTURE", sources%apertureRadius)

    call loadInt8Column(inTableH, "EFLAG", sources%extensionFlag)

  
    sources%innerBckRadius = innerBackgroundRadius

    sources%outerBckRadius = outerBackgroundRadius
    sources%exposure = exposure 

   
 
  
    do i = 1, nsources
        sources(i)%number = i
    end do


    call release(set)






    allocate(oldsources(1 : nSources) )  ! Stores the old source-list




    oldSources = sources

    fastBackground = .false.

end subroutine loadSources
!*******************************************************************************
!
!*******************************************************************************
subroutine recompute_photometry(numberToFit, no, type, radius, &
      innerA, outerA, axialRatio, theta, background)
    integer, intent(in) :: numberToFit
    integer, intent(in) :: no(1 : numberToFit)

    
    integer, intent(in) :: type
    real(double), intent(in) :: radius, innerA, outerA, axialRatio, theta, background

    real(single) :: corrBackgroundCountRateErr = 0.0
    type(SourceParameters), pointer  :: temp(:) => null()
    type(SourceParameters) :: s
    integer(int8) :: i


    ! Do aperture photometry on  source

    allocate(temp( 1 : numberToFit))

    do i = 1, numberToFit

!write(*, *) i, ") no=", no(i), " aperture radius=", radius, " innera=", innera, " outer=", outera, &
! " sourceno=", oldSources(no(i))
    
        temp(i) = oldSources(no(i))

        temp(i)%apertureradius = radius
        temp(i)%innerbckradius = innerA
        temp(i)%outerbckradius = outerA
        temp(i)%exposure = exposure
        temp(i)%axialRatio = axialRatio
        temp(i)%theta = theta
        temp(i)%rawCountRate = -1.0
        temp(i)%rawBackgroundCountRate = -1.0

        temp(i)%exposure = exposure

    end do     


    if (.not. associated(myImage)) then
        call fatal("myimage not associated")

    end if


    if (.not. associated(mylevelImage)) then
        call fatal("mylevelimage not associated")

    end if


    if (.not. associated(myqualityImage)) then
        call fatal("myqualityimage not associated")

    end if


    !************************************************************************************
    ! Point-source photometry if type equals 1
    !************************************************************************************ 
    select case (type)




        case(-2)  ! Extended-source photometry - fixed background
             temp(1)%extensionFlag = 1
             temp(1)%rawBackgroundCountRate = background
             write(errstr,'(a, f8.1)') "Performing extended-source photometry- "&
             // "background counts per pixel=", background
            call message(AppMsg, Verbose)


            !************************************************************************************
            ! Extended photometry
            !************************************************************************************ 
            call ExtendedSourcePhotometry( myimage, temp(1), myApertureRadius / real(binx), &
                     temp(1)%apertureRadius, temp(1)%apertureRadius / temp(1)%axialRatio, temp(1)%theta)



        case(-1)  ! Extended-source photometry
             temp(1)%extensionFlag = 1

            write(errstr,'(a)') "Performing extended-source photometry- "&
             // "background subtracted by interpolation using pixels within the background annulus"
            call message(AppMsg, Verbose)


            !************************************************************************************
            ! Extended photometry
            !************************************************************************************ 
            call ExtendedSourcePhotometry( myimage, mybackgroundimage, temp(1), myApertureRadius / real(binx), &
                     temp(1)%apertureRadius, temp(1)%apertureRadius / temp(1)%axialRatio, temp(1)%theta)





        case (0)  ! DEFAULT APERTURE PHOTOMTERY
            temp(1)%extensionFlag = 0



            ! Aperture photometry
            write(errstr,'(a)') "Computing photometry using circular aperture"
            call message(AppMsg, Verbose)



            if( fastBackground ) then

                !*************************************************************
                ! For FAST data might be using a bigger image to get the
                ! background
                !*************************************************************
                call getFastBackground( temp(1))
    

                call AperturePhotometry2(int(nx, int16), int(ny, int16), &
                    myImage, mylevelImage,                                 &
                    myQualityImage, myapertureRadius / real(binX),         &  
                    1_int16, temp)
        
            else    

                                  

!                write(*, *) "Aperture photometry line 591, my apertureradius=", myApertureRadius, &
!                   " source aperture radius=", temp(1)%apertureradius

                                 

                call AperturePhotometry1(int(nx, int16), int(ny, int16), &
                    myImage, myBackgroundImage, mylevelImage,          &
                    myQualityImage, myapertureRadius / real(binX),     &  
                    1_int16, temp)

!                write(*, *) "x=", temp(1)%xpos, " y=", temp(1)%ypos, " apertureradius=", temp(1)%apertureradius
!                write(*, *) "Exposure=", temp(1)%exposure
!                write(*, *) "Computed raw countrate=", temp(1)%rawCountRate
!                write(*, *) "Computed raw backgroundcountrate=", temp(1)%rawbackgroundCountRate
!                write(*, *) "Inner background=", temp(1)%innerbckradius
!                write(*, *) "outer background=", temp(1)%outerbckradius
  
  

            end if 


        case (1)  ! DEFAULT APERTURE PHOTOMTERY - fixed background
            temp(1)%extensionFlag = 0
            temp(1)%rawBackgroundCountRate = background
   


            ! Aperture photometry
            write(errstr,'(a, f6.1)') &
                 "Computing photometry using circular aperture- background counts per pixel=", background
            call message(AppMsg, Verbose)

          
        
            call AperturePhotometry2(nx, ny, myImage, mylevelImage,                           &
                    myQualityImage, myapertureRadius / real(binX),                           &  
                    1_int16, temp)
        
!             write(*, *) "Line 579 rawcountrate=", temp(1)%rawCountRate
!              write(*, *) "Line 579 corrcountrate=", temp(1)%corrCountRate


!             write(*, *) "Line 579 rawbackgroundcountrate=", temp(1)%rawBackgroundCountRate
!             write(*, *) "Line 579 corrbackgroundcountrate=", temp(1)%corrBackgroundCountRate
!             write(*, *) "Total raw counts=", (temp(1)%rawBackgroundCountRate +  temp(1)%rawCountRate) * temp(1)%exposure
                  
        case(2)   ! Least-squares fit using background from annulus

            temp(1)%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from a least-squares fit to the PSF which excludes a constant term -" &
                // " background from annulus"
            call message(AppMsg, Verbose)





            call PSFPhotometry1(myimage, temp(1), real(innerA), real(outerA))


           
        case(3)   ! Least-squares to PSF to one or more sources, including background

      

            temp%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from a least-squares fit to the PSF which includes " &
                // "a constant term for the background"
            call message(AppMsg, Verbose)

            if(numberToFit > 1) then

                write(errstr,'(a, i3, a, 10i6)') "Number of sources included in the fit=", numberToFit, &
                 " source nos=", no(1 : numberToFit) 
                call message(AppMsg, Verbose)
            end if 


            call PSFPhotometry3(int(numberToFit, int8), myimage, temp)

 



        case(4)   ! Least-squares fit to several sources excluding background



            temp%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a, f6.1)') "Computing photometry from a least-squares fit to the PSF with a fixed background- ", &
                background
            call message(AppMsg, Verbose)

            if(numberToFit > 1) then

                write(errstr,'(a, i3, a, 10i6)') "Number of sources included in the fit=", numberToFit, &
                    " source nos=", no(1 : numberToFit) 
                call message(AppMsg, Verbose)
            end if 


            call PSFPhotometry33(int(numberToFit, int8), myimage, temp, real(background))

 

        case(5)  ! least-squares fit including fit to background and position


            temp(1)%extensionFlag = 0


            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from a least-squares fit to the PSF which "&
              // "includes a constant term for the background"
            call message(AppMsg, Verbose)

            write(errstr,'(a)') "Coords recomputed from a non-linear least-squares fit"
            call message(AppMsg, Verbose)


            if(numberToFit > 1) then

                write(errstr,'(a, i3, a, 10i6)') "Number of sources included in the fit=", numberToFit, &
                    " source nos=", no(1 : numberToFit) 
                call message(AppMsg, Verbose)
            end if 


            call PSFPhotometry4(myimage, temp, numberToFit)


        case(6)  ! least-squares fit including fit to position - fixed background


            temp(1)%extensionFlag = 0

            write(errstr,'(a)') "Computing photometry from a least-squares fit to the PSF which "&
              // "includes a constant term for the background"


            call message(AppMsg, Verbose)


            write(errstr,'(a)') "Coords recomputed from a non-linear least-squares fit"

            call message(AppMsg, Verbose)

            if(numberToFit > 1) then

                write(errstr,'(a, i3, a, 10i6)') "Number of sources included in the fit=", numberToFit, &
                   " source nos=", no(1 : numberToFit) 
                call message(AppMsg, Verbose)
            end if 


            call PSFPhotometry44(myimage, temp, numberToFit, real(background))



        case default
            call fatal("UNKNOWN PHOTOMETRY TYPE!")
    end select  




    !************************************************************************************
    ! Compute the magnitude
    !
    !************************************************************************************

    
    s = temp(1)  

    call ComputeMagnitude(s%corrCountRate, s%corrCountRateErr,                   &
                          s%corrBackgroundCountRate, corrBackgroundCountRateErr, &
                          s%xPos, s%yPos,                                        &
                          myScaleCounts, s%extensionFlag,                        &
                          myCalRadius, myApertureRadius,                         &
                          frameTime, deadFrac,                                   &
                          s%magnitude, s%magnitudeErr, s%psf2Correction,         &
                          filter, mjd, s%tdsCorr)



!write(*, *) "Computed magnitude=", s%magnitude
!write(*, *) "mjd=", mjd, " tdscorr=", s%tdsCorr
    !************************************************************************
    ! Set the source in the sourcelist to the new one
    !***********************************************************************  
    sources(no(1)) = s

    

    deallocate( temp )
    nullify( temp )

end subroutine recompute_photometry


!*******************************************************************************
!
!*******************************************************************************

subroutine getoldsource( no, oldSource )
    integer, intent(in) :: no
    type(SourceParameters), intent(out) :: oldSource


    oldsource = oldsources(no)
    

end subroutine getOldSource


!*******************************************************************************
!
!*******************************************************************************

subroutine getnewsource( no, newSource )
    integer, intent(in) :: no
    type(SourceParameters), intent(out) :: newSource


    newsource = sources(no)
    

end subroutine getnewSource






!*******************************************************************************
!
!*******************************************************************************

subroutine finish( )
    type(dataSetT) :: dataSetH   ! For creating the new source-list
    type(TableT)  :: listTable
   




!write(*, *) "omphotom_f1 Finish line 850" 






!write(*, *) "Backgroundpixel is ", associated(backgroundpixel)
!write(*, *) "myImage is ", associated(myImage)
!write(*, *) "mybackgroundimage is ", associated(mybackgroundImage)
!write(*, *) "mylevelimage is ", associated(mylevelImage)
!write(*, *) "myqualitylimage is ", associated(myqualityImage)
!write(*, *) "usebackgroundimage=", usebackgroundimage








    ! Write out the source list to the oswDataSet
    write(errstr,'(a)') "Creating new source list"
    call message(AppMsg, Verbose)




    !***********************************************************
    ! Create a copy of the old source-list file 
    !***********************************************************

    call copydataSet( trim(stringParameter("oldsrclistset")) ,  trim(stringParameter("newsrclistset")))


    !**********************************
    ! Create the output source list
    !**********************************

    dataSetH = dataSet(trim(stringParameter("newsrclistset")), modify)





    listTable = Table(dataSetH, "SRCLIST")



  
    call AddInt8Column(listTable, sources%extensionFlag, nSources, &
              "EFLAG", "none", "0=point, 1=extended source")    ! Can be either 0 (point-source) or 1 (extended)
 
      

    call AddReal32Column(listTable, sources%rawCountRate, nSources, &
                        "RATE", "counts/sec", "Source-background. No corrections have been applied", nullValue)



    call AddReal32Column(listTable, sources%rawCountRateErr, nSources, &
                        "RATE_ERR", "counts/sec", "No corrections have been applied", nullValue)
  

    call AddReal32Column(listTable, sources%rawBackGroundCountRate, nSources, &
              "BACKGROUND_RATE", "counts/sec",  "No corrections have been applied", nullValue)
    


    call AddReal32Column(listTable, sources%significance,  nSources, &
               "SIGNIFICANCE", "sigma", "(source-background)/sqrt(source+background)", nullValue)
 


    ! Add the coincidence-corrected source-count rates 
 
    call AddReal32Column(listTable,  sources%corrCountRate, nSources, &
                "CORR_RATE", "counts/sec", "Source-background. Corrected for coincidence losses, etc", nullValue)



  
    call AddReal32Column(listTable,  sources%corrCountRateErr, nSources, &
                "CORR_RATE_ERR", "counts/sec", "Corrected for coincidence losses", nullValue)
    


    call AddReal32Column(listTable, sources%corrBackGroundCountRate, nSources, &
                         "CORR_BACKGROUND_RATE", "counts/sec", "Corrected for coincidence losses, etc", nullValue)


    ! Add the coincidence-loss correction  
    call AddReal32Column(listTable, sources%psf1Correction, nSources, &
              "PSF1_CORR", "%", "PSF correction from mesurement to 12 pixel aperture", nullValue)



    call AddReal32Column(listTable, real(sources%rightAscension), nSources, &
              "SPB_COILOSS", " ", "(S+B coinc-loss corr)/(S+B un-coinc-loss corr)", nullValue)


    call AddReal32Column(listTable, real(sources%declination), nSources, &
              "BK_COILOSS", " ", "(B coinc-loss corr)/(B un-coinc-loss corr)", nullValue)

 

    ! Add the aperture radius 
    call AddReal32Column(listTable, sources%apertureRadius, nSources, &
                    "APERTURE", "unbinned pixels", "Aperture radius used in photometry", nullValue)



    ! Add the magnitudes 
    call AddReal32Column(listTable, sources%magnitude, nSources, &
                    "MAG", "", "Instrumental mags", nullValue)


    ! Add the magnitudes 
    call AddReal32Column(listTable, sources%magnitudeErr, nSources, &
                    "MAGERR", "", "Instrumental mag errors", nullValue)



    call release(dataSetH)


    !********************************************************************
    ! Delete arrays
    !*******************************************************************


    if( useBackgroundImage ) then ! The background image was not associated with the image so delete it 

    


        if(associated( myBackgroundImage )) then

            deallocate( myBackgroundImage )
           
        end if 
    end if




    !***********************************************
    ! Nullify the background image
    !**********************************************

    nullify( myBackgroundImage )


 

    if(associated( myImage )) then 
        deallocate( myImage )
        nullify( myImage )
    end if



    if(associated(myLevelImage)) then
        deallocate( myLevelImage )
        nullify( myLevelImage )

    end if


 

    if(associated( myQualityImage )) then
        deallocate( myQualityImage )
        nullify( myQualityImage )

    end if


    if(associated( sources )) then
        deallocate( sources )
        nullify( sources )

    end if


    if(associated( oldsources )) then
        deallocate( oldsources )
        nullify( oldsources )
    end if
  

end subroutine finish

!**************************************************************************
! 
! For extended sources with elliptical apertures
! semia = semi-major axis of aperture
! semib = semi-minor axis of aperture
! angle is the position angle of the ellipse
! source%innerBckRadius = semi-major axis of inner background annulus
! source%outerBckRadius = semi-major axis of outer background annulus
! 
!**************************************************************************
subroutine ExtendedSourcePhotometry1( image, backgroundImage, &
                source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer, dimension(:, :) :: image, backgroundimage
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta, thetar
      
    real(double) :: f, fb, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea, coincLossArea
 
    real(single), pointer, dimension(:) :: x=>null(), y=>null(), counts=>null()
    integer :: ncount
          

    integer, parameter :: maxSize = 10000
    real(single) :: innerbckSemia, outerbcksemia
    real(single) :: innerbckSemib, outerbcksemib
    integer :: imin, imax, jmin, jmax


    coincLossArea = PI * coincLossRadius * coincLossRadius

    ccdImage => image
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    thetar = theta * degreesToRadians


    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
           
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     


    !***************************************************************************
    ! Set up a background interpolant using the pixels in the background annulus
    !************************************************************************** 

    x => allocateSingle( 1, maxSize, "x")
    y => allocateSingle( 1, maxSize, "y")
    counts => allocateSingle( 1, maxSize, "counts")


    innerbcksemia = source%innerBckRadius
    outerbckSemia = source%outerBckRadius
    innerbcksemib = source%innerBckRadius / source%axialRatio
    outerbcksemib = source%outerBckRadius / source%axialRatio

    ncount = 0
    imin =  max(0, int(floor(source%xPos) - ceiling(outerbcksemiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + outerbcksemiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(outerbcksemiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + outerbcksemiA + 0.1) + 1) 
 


iloop :    do i = imin, imax 
        x0 = real(i + 1)  
    
        do j = jmin, jmax
            if(myQualityImage(i, j)) cycle  ! Don't use bad pixels
                                   
            y0 = real(j + 1) 
      
            !********************************************************************
            ! Point has to be within the inner and outer ellipse
            !********************************************************************
            if (.not. PointInEllipse(x0, y0, source%xpos, source%ypos, innerbcksemiA, innerbcksemiB, thetar) .and. &
                      PointInEllipse(x0, y0, source%xpos, source%ypos, outerbcksemiA, outerbcksemiB, thetar)) then

           
                                 
                ncount = ncount + 1
                x(ncount) = real(i) + 0.5
                y(ncount) = real(j) + 0.5
                    
                counts(ncount) = backgroundimage(i, j)  
     
 !              image(i, j) = -1.0          
    
                if (ncount == maxSize) then
                      
                  exit iLoop
                end if  
            end if
                    
        end do    
    end do iLoop

 

   
    backgroundCounts = 0.0_double
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
 


    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 



    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
       
            if(myQualityImage(i, j)) cycle  ! Don't use bad pixels
             
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea

            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 

            !***************************************************************************
            ! Interpolate to get the background counts
            !**************************************************************************
            fb = neighbourInterpolate( x0, y0, ncount, x, y, counts, 4.0 * semiB ) 
          

            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
                fb = fb * pixarea 
            else
                f =  PixelValue2(i, j)            
              
            end if
            backgroundCounts = backgroundCounts + fb
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), frameTime)
          
!image(i, j) = fb                       
        end do
    end do


!call SaveImage(image, "omsourceimage.fits", "test")








!write(*, *) "OmsourceExtendedSourceCounts2 semiA=", semia, " semiB=", semib, "angle=", angle, " total area=", totalarea
  
!write(*, *) "exposure=", exposure, " coinclossarea=", coinclossarea 


    backgroundRate = coincLossArea * backgroundCounts / (totalArea * source%exposure) ! Background count rate

!write(*, *) "background counts=", backgroundcounts, " backgroundrate=", backgroundrate


    ! The raw background count-rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      


    !******************************************************************
    ! The coincidence-loss correction for the background count-rate
    !******************************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)

!write(*, *) "Coincidence-loss correction for background=", correction
!write(*, *) "background counts=", backgroundcounts, " total area=", totalarea, " coinclossarea=", coinclossarea
    
    ! The coincidence-loss corrected background count-rate-rate 
    source%corrBackGroundCountRate = source%rawBackgroundCountRate * correction ! Corrected background rate




    correctedBackgroundCounts = backgroundCounts * correction 

!write(*, *) "corrected total counts=", correctedcounts, " correctedbackgroundcounts=", correctedBackgroundCounts 


    !****************************************************
    ! The source coincidence-loss correction
    !****************************************************
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  correctedCounts / rawCounts
    end if

    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  correction
    end if



    !********************************************************
    ! The background subtracted raw count-rate
    !*******************************************************
    rawCounts = rawCounts - backgroundCounts  
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               

    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts)  / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 

      
    source%psf1Correction = 0.0


    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
 
    nullify( ccdImage )
    deallocate(x, y, counts)
    nullify(x, y, counts)

end subroutine extendedSourcePhotometry1      



!**************************************************************************
! 
! For extended sources with elliptical apertures
! semia = semi-major axis of aperture
! semib = semi-minor axis of aperture
! angle is the position angle of the ellipse
! source%innerBckRadius = semi-major axis of inner background annulus
! source%outerBckRadius = semi-major axis of outer background annulus
! 
!**************************************************************************
subroutine ExtendedSourcePhotometry2( image, &
                source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer, dimension(:, :) :: image
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta, thetar
      
    real(double) :: f, fb, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea, coincLossArea
 
 
    integer :: ncount
          

    integer, parameter :: maxSize = 10000
    real(single) :: innerbckSemia, outerbcksemia
    real(single) :: innerbckSemib, outerbcksemib
    real(single) :: theBackground
    integer :: imin, imax, jmin, jmax


    coincLossArea = PI * coincLossRadius * coincLossRadius

    ccdImage => image
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    thetar = theta * degreesToRadians


    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
           
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     


    !***************************************************************************
    ! Set up a background interpolant using the pixels in the background annulus
    !************************************************************************** 
 
    innerbcksemia = source%innerBckRadius
    outerbckSemia = source%outerBckRadius
    innerbcksemib = source%innerBckRadius / source%axialRatio
    outerbcksemib = source%outerBckRadius / source%axialRatio

    ncount = 0
    imin =  max(0, int(floor(source%xPos) - ceiling(outerbcksemiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + outerbcksemiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(outerbcksemiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + outerbcksemiA + 0.1) + 1) 

   
    theBackground = source%rawBackgroundCountRate
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  

    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 

    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
       
            if(myQualityImage(i, j)) cycle  ! Don't use bad pixels
                  
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea

            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 

            !***************************************************************************
            ! Use the given background counts/pixel
            !**************************************************************************
            fb = theBackground 
          

            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
                fb = fb * pixarea 
            else
                f =  PixelValue2(i, j)            
              
            end if
            backgroundCounts = backgroundCounts + fb
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), frameTime)
          
!image(i, j) = fb                       
        end do
    end do


!call SaveImage(image, "omsourceimage.fits", "test")








!write(*, *) "OmsourceExtendedSourceCounts2 semiA=", semia, " semiB=", semib, "angle=", angle, " total area=", totalarea
  
!write(*, *) "exposure=", exposure, " coinclossarea=", coinclossarea 


    backgroundRate = coincLossArea * backgroundCounts / (totalArea * source%exposure) ! Background count rate

!write(*, *) "background counts=", backgroundcounts, " backgroundrate=", backgroundrate


    ! The raw background count-rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      


    !******************************************************************
    ! The coincidence-loss correction for the background count-rate
    !******************************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)

!write(*, *) "Coincidence-loss correction for background=", correction
!write(*, *) "background counts=", backgroundcounts, " total area=", totalarea, " coinclossarea=", coinclossarea
    
    ! The coincidence-loss corrected background count-rate-rate 
    source%corrBackGroundCountRate = source%rawBackgroundCountRate * correction ! Corrected background rate




    correctedBackgroundCounts = backgroundCounts * correction 

!write(*, *) "corrected total counts=", correctedcounts, " correctedbackgroundcounts=", correctedBackgroundCounts 


    !****************************************************
    ! The source coincidence-loss correction
    !****************************************************
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  correctedCounts / rawCounts
    end if

    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  correction
    end if



    !********************************************************
    ! The background subtracted raw count-rate
    !*******************************************************
    rawCounts = rawCounts - backgroundCounts  
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               

    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts)  / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 

      
    source%psf1Correction = 0.0


    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
 
    nullify( ccdImage )
    

end subroutine extendedSourcePhotometry2      



!*********************************************
! Subroutine GetImageSize
!********************************************
subroutine GetImageSize(oswDataSet, windowX0,  windowY0,  windowDX, windowDY)
    type(DataSetT), intent(in) :: oswDataSet 
    integer(int32), intent(out) ::  windowX0,  windowY0,  windowDX, windowDY
    
    type(TableT) :: inTableH
    integer, parameter :: noColumns = 4
    type(ColumnT) :: inColumnH(noColumns)
    integer(int32) :: i
    integer(int32), pointer ::  windowX0P(:),  windowY0P(:),  windowDXP(:), windowDYP(:)
   
    

    if (hasAttribute(oswDataSet, "WINDOWX0") .and. &
        hasAttribute(oswDataSet, "WINDOWY0") .and. &
        hasAttribute(oswDataSet, "WINDOWDX") .and. &
        hasAttribute(oswDataSet, "WINDOWDY")) then

        windowx0 = int32Attribute(oswdataset, "WINDOWX0")
        windowy0 = int32Attribute(oswdataset, "WINDOWY0")
        windowdx = int32Attribute(oswdataset, "WINDOWDX")
        windowdY = int32Attribute(oswdataset, "WINDOWDY")
        return
    end if 

    !**************************************************
    ! Get a handle on the modes binary table extension.
    !**************************************************


    inTableH = table(oswDataSet, "MODES")

    !**************************************************
    ! Get handles and pointers on the modes binary table 
    ! extension columns and extract the window data.
    !**************************************************
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)


    !************************************
    ! Release the handles
    !***********************************
    do i = 1, noColumns
        call release(inColumnH(i))
    end do
    call release(inTableH)

end subroutine GetImageSize


!***************************************************************
! Gets the background from the background image
!
!***************************************************************
subroutine getFastBackground( source  )

    type(SourceParameters), intent(inout) :: source
 
    real(single) :: xPos, yPos, mode, error

    !******************************************************************
    ! Get the full-frame image coordinates of the source
    !******************************************************************
    xPos = source%xpos
    yPos = source%ypos



    call ConvertCoordsTofullFrame( xPos, yPos, X0image, Y0image, imageBinX, imageBinY)



    !*************************************************************************
    ! Convert to x, y image coordinates of background image
    !*************************************************************************           

    call ConvertCoordsFromfullFrame( xPos, yPos, X0Background, Y0Background, &
                                     backgroundBinX, backgroundBinY)
 

    call initialiseArrays


    ! Write out the source list to the oswDataSet
    write(errstr,'(a, a)') "Obtaining background from ", trim(stringParameter("backgroundimageset")) 
    call message(AppMsg, Verbose)


    !*************************************************************
    ! Get the background mode
    !*************************************************************
    mode = GetBackGroundMode4(  &
        mybackgroundImage,      &  
        xpos - 0.5,             &   ! The centre x-pixel coordinate of the aperture
        ypos - 0.5,             &   ! The centre y-pixel corrdinate of the aperture 
        source%innerBckRadius,  &   ! The aperture inner radius
        source%outerBckRadius,  &   ! The aperture outer radius
        error)      ! The error

    call deleteArrays

    source%rawBackgroundCountRate = mode

    write(errstr,'(a, f6.2)') "Computed background mode=", mode
    call message(AppMsg, Verbose)




end subroutine getFastBackground


!***************************************************************
! For the FAST-MODE if an image from the IMAGING chain is given
! to get the background, then need to make the image the
! same sze
!***************************************************************
subroutine getParameters
  
    
    type(DataSetT) :: dataSetH
    integer :: dx, dy 
    logical :: binBpe
     
    !***********************************************
    ! Get the window parameters for the image 
    !**********************************************
    dataSetH = dataSet( trim(stringParameter("imageset")), read)
    call GetImageSize( dataSetH, X0Image,  Y0Image,  dx, dy)
    call release(dataSetH)

    imageBinx = binx

    imageBiny = biny

    !*****************************************************
    ! Get the window parameters for the background image 
    !******************************************************

    dataSetH = dataSet( trim(stringParameter("backgroundimageset")), read)
    call GetImageSize( dataSetH, X0Background,  Y0Background,  dx, dy)


    ! compute binning 
    binBpe = booleanAttribute(dataSetH, "BINBPE")
    backgroundBinx = 2 ** int32Attribute(dataSetH, "BINAX1")
    backgroundBiny = 2 ** int32Attribute(dataSetH, "BINAX2")


    if (binBpe) then
        backgroundBinx = backgroundBinx * 2
        backgroundBiny = backgroundBiny * 2
    end if
    
    call release(dataSetH)


end subroutine getParameters


end module omphotom_f1










