!************************************************************
! This routine contains routines for doing photometry
! using the PSF 
!
!***********************************************************
module psfphotometry_mod
use dal
use cal
use sasdal
use oal
use errorHandling
use frametime_mod, only : getFrameTime ! mssllib
use getFilterID_mod
use omutility_mod      ! In mssllib various fits access routines
use omconstants_mod    ! In mssllib defines various constants
use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
use omparameters_mod 
use omphotometry_mod   ! In mssllib for doing aperture photometry
use ommagnitude_mod    ! In mssllib for computing instrumental magnitude 
use ominterpolate_mod
use omnumerics_mod
use ompsf_mod
use leastsquares_mod
use omminim_mod
use omsourceflags_mod
implicit none
 
!********************************************************************
! Global variables
!********************************************************************




    real(single), private :: myApertureRadius = 0.0, myCalRadius = 0.0




    real(single), private, pointer :: myImage(:, :) => null()

    real(single), private, pointer :: myBackgroundImage(:, :) => null()


  

    integer(int16), private, pointer :: mylevelImage(:, :) => null()

    logical(boolean), private, pointer :: myQualityImage(:, :) => null()

 



    real(single), private, pointer :: mypsfImage(:, :) => null()

    
    real(single), private :: meanBackground


  

    !*******************************************************************************
    ! The following are for PSF photometry
    !*******************************************************************************

    integer(int8), private, parameter :: maxNoSources = 5
    integer, private :: imin, imax, jmin, jmax
    real(single), private , pointer, dimension(:, :) :: xdata => null()
    real(single), private,  pointer, dimension(:) :: ydata => null(),  &
           weight => null()

    integer(int16), private :: l0=0


    real(single), private :: psfxCentre, psfYcentre, sumPsfImage



    real(double), private :: bl(10), bu(10)
    integer(int8), parameter :: maxShift = 4 

    integer, private :: psfNx, psfny
    integer(int8) :: numberToFit

    logical :: fastBackground = .false.







   real(single), private :: exposure = 0.0
    real(double), private :: plateScale 
    real(double), private :: mjd = 0.0
    character(10), private :: myFilter     



    logical, private  :: myScaleCounts = .false.
    real(double), private ::  myScaleCountsRatio = 1.0 


interface initialisepsfphotometry 

    module procedure initialisepsfphotometry1, initialisepsfphotometry2
end interface initialisepsfphotometry


contains


!*********************************************************************
! Info: Initialises the fortran part of omsource.
!       It loads the image pdata and properties, and -if any- it also
!       creates a source list from an already existing sources dataset
!*********************************************************************
subroutine initialisePSFphotometry1(imageFile)
    character(*), intent(in) :: imageFile
    
    real(single) :: sigma 
       
    type(dataSetT) :: set
    logical :: binBpe = .false.
 






    if(associated(myImage)) then
        deallocate(myImage)
        nullify(myImage)
    end if





    if(associated(myBackgroundImage)) then
        deallocate(myBackgroundImage)
        nullify(myBackgroundImage)
    end if



    if(associated(mylevelImage)) then
        deallocate(myLevelImage)
        nullify(myLevelImage)
    end if



    if(associated(myQualityImage)) then
        deallocate(myQualityImage)
        nullify(myQualityImage)
    end if

   




    ! --- extract the image from the file
    set = dataSet( trim(imageFile), read)

     
    exposure = real32Attribute( set, "EXPOSURE")  
    myfilter = stringAttribute( set, "FILTER" )   

    ! compute binning
    binBpe = booleanAttribute(set, "BINBPE")
    binx     = 2 ** int32Attribute(set, "BINAX1")
    biny     = 2 ** int32Attribute(set, "BINAX2")


    if (binBpe) then
        binx= binx * 2
        biny= biny * 2
    end if
    


    ! CCD frame-time and dead-fraction
   
          
    frameTime = real64Attribute(set, "FRAMTIME") / 1000.0
    deadFrac = real64Attribute(set, "DEADFRAC")






    !*************************************************************************
    ! Need the modified Julian date for time-dependent sensitivity correction
    !************************************************************************
     mjd = ModifiedJulianDate( trim( stringAttribute(set, "DATE-OBS")))


    call CAL_setState(OM, filterId = filterIdToInt(trim( myfilter ) ))

    call setSignificanceRadius( myfilter ) 
      

      ! now init radii once, and once only, and bin them now
    myApertureRadius = 12.0                ! NEED THIS IN UNBINND PIXELS

    myCalRadius = CAL_omGetApertureRadius() ! NEED THIS IN UNBINNED PIXELS


    if (myCalRadius .gt. myApertureRadius) then
        !**********************************************
        ! Remember that counts have to be extrapolated.
        !**********************************************
        myScaleCounts = .true.
        myScaleCountsRatio = (myCalRadius / myApertureRadius) ** 2 
    end if



    !************************************* 
    ! Get the Plate scale (arcsecs/pixel)
    !*************************************
    call CAL_omGetPlateScale(0.55_double, plateScale)

    ! Get the pixels
    call createImageArray( set, myImage )



    ! Get the quality of the pixels
!    call createQualityArray( set, myQualityImage )
   

    if(associated(myQualityImage)) then
        deallocate(myQualityImage)
        nullify(myQualityIMage)
    endif

    nx = size(myImage, 1)
    ny = size(myImage, 2)
 



    myQualityImage => getQualityArray(set, nx, ny)


    call release( set )  




    ! Compute the background mode

   
 

    
    meanBackground = backgroundMode(myImage, int(nx, int16), int(ny, int16), 2.0, sigma) 


!    myLevelImage => allocate2dInt16(0, nx - 1, 0, ny - 1, "myLevelImage")


  

    myBackgroundImage => myImage

    fastBackground = .false.  

    









    !*****************************************************************
    !
    !*****************************************************************
    call initialisePSFPhotometry(myimage, myfilter, myapertureRadius, mycalRadius)

 



    

end subroutine initialisePsfphotometry1







!*****************************************************************
!
!*****************************************************************
subroutine initialisePSFPhotometry2(image, filter, apertureRadius, calRadius)

    real(single), pointer :: image(:, :)
    character(*), intent(in) :: filter
    real(single), intent(in) ::  apertureRadius, calRadius


    myImage => image

    myFilter = trim(filter)


    myApertureRadius = apertureRadius 
    myCalRadius = calRadius




    !**************************************************
    ! Create the psf image
    !*************************************************
!    if( .not. associated( myPsfImage) ) then
    call CreatethePsfImage
!    end if





end subroutine initialisePSFPhotometry2



!*****************************************************
!
!
!*****************************************************
subroutine CreatethePsfImage
    real(double) :: maxPsf
   
    real(single) ::  counts
    integer :: oldNx, oldNy

    integer i

    real(single), pointer, dimension(:) :: x=>null(), y=>null()




      




    oldnx = nx
    oldny = ny

   
    psfnx = 12
    psfny = 12
       
     

    l0 = 14 / binx

    if(associated( myPsfImage )) then
        deallocate( myPsfImage )
        nullify( myPsfImage )
    end if



    ! Get the point-spread function image
  
    call CreatePSFImage21(psfNx, psfNy, mypsfImage, 1024.5_double, 1024.5_double, &
        ratio=0.5_double, bin=binX, res=1) 




!    write(*, *) "Counts=", counts, " area=", area, " total counts from image", sum(mypsfImage)


    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  


    psfXcentre = real(psfnx) / 2.0 ! The x centre of the psf image
    psfYcentre = real(psfny) / 2.0 ! The y centre of the psf image




    !*******************************************************
    ! Normalize psf image to a maximum height of 1.0
    !*******************************************************

    maxPsf = maxval(myPsfImage) 
    mypsfImage = mypsfImage / maxPsf


    call GetCounts2( mypsfImage, psfXCentre, psfYCentre, real(12 / binx), counts)

    sumPsfImage = counts


    x => allocateSingle(1, psfnx, "x")
    y => allocateSingle(1, psfnx, "y")

    do i = 1, psfnx
        x(i) = real(i) - 0.5
        y(i) = real(i) - 0.5

    end do


    !*************************************************************************
    !
    !*************************************************************************
    call gridInterpolantConstruct(psfnx, psfny, x, y, myPsfImage, psfgridfit)

!x = x + 0.5
!y = y + 0.5
!do i = 1, psfnx 
!do j = 1, psfny 
!write(*, *) "pixel ", i, ", ", j, " x=", x(i), " y=", y(j), " image=", myPsfImage(i-1, j-1), &
!" interpolated=", gridInterpolate(x(i), y(j), psfgridfit), " bilinear interpolated=", &
!bilinearInterpolate(mypsfimage, x(i), y(j))
!end do
!end do

!    call gridInterpolantDestruct( psfgridFit )

    deallocate(x, y)
    nullify(x, y)
    nx = oldnx
    ny = oldny    


    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1 





 end subroutine CreatethePsfImage

!*************************************************************************
! Least-squares fit to the PSF, not including the background
!
!
!************************************************************************

subroutine PSFPhotometry1(image, source, innerA, outerA)
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters), intent(inout) :: source
    real(single), intent(in) :: innerA, outerA 

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, imin, imax, jmin, jmax, n
    real(single), pointer, dimension(:) :: xdata => null(), ydata => null(), weight => null()
  
    real(single) :: xpos, ypos, dx, dy

  
    real(single) :: gradient, gradientError
 
     
    real(single) :: mode, error







    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    imin = max(0, i0 - l0)
    imax = min(nx, i0 + l0)

    jmin = max(0, j0 - l0)
    jmax = min(ny, j0 + l0)
   

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocateSingle(1, n, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
  



 
    if( .not. fastBackground) then
        call initialiseArrays
        ccdImage => image




        !*************************************************************
        !
        !*************************************************************
        mode = GetBackGroundMode3(  &
        source%xpos - 0.5,          &   ! The centre x-pixel coordinate of the aperture
        source%ypos - 0.5,          &   ! The centre y-pixel corrdinate of the aperture 
        innerA, &   ! The aperture inner radius
        outerA, &   ! The aperture outer radius
        error)      ! The error
        call deleteArrays
        nullify(ccDImage)


    else
        mode = source%rawBackgroundCountRate
    end if




    do i = imin, imax

        dx = real(i + 1) - source%xpos  ! NB (1, 1) corresponds to pixel 0, 0

        xpos = psfXcentre + dx 
  
        if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
        do j = jmin, jmax
          
           
 
            dy = real(j + 1) - source%yPos 
     
            ypos = psfYcentre + dy
  
            if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
            n = n + 1
           
            !********************************************************
            ! NB pixel 0, 0 has coordinates 0.5, 0.5
            !********************************************************
       

            xData(n) = PsfInterpolation(xPos, yPos)

            if (xData(n) .lt. 0.0) then
                n = max(0, n - 1)
                cycle
            end if

            ydata(n) =  image(i, j)


!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


           
            weight(n) = weightingFunction( ydata(n)) 
            

        end do
    end do






  
    !************************************************************************
    ! Least-squares fit with no constant term
    !************************************************************************   

    call LeastSquaresFit1(n, xdata, ydata, weight, gradient, gradientError)


    gradient = gradient * sumPsfImage

  

    source%apertureRadius = myApertureRadius
 
    source%rawCountRate = gradient / source%exposure
    source%rawBackgroundCountRate = mode * PI * real(144  / &
                  (binX * binX)) / source%exposure



    call PointSourceCoincLossCorr( source, myApertureRadius, frameTime )     



!  write(*, *) "Original raw counts =", source%rawCountRate * source%exposure, & 
!" ratio (new/old)=",  (gradient / (source%rawCountRate * source%exposure) )
 

    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)





    
end subroutine PSFPhotometry1

!*************************************************************************
!
! Least-squares to the PSF to include the background
!
!************************************************************************

subroutine PSFPhotometry2(image, source)
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters), intent(inout) :: source
   

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, imin, imax, jmin, jmax, n
    
  
    real(single) :: xpos, ypos, dx, dy

  
    real(single) :: gradient(1), gradientError(1), constant, constantError
    
   
 
    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    imin = max(0, i0 - l0)
    imax = min(nx, i0 + l0)

    jmin = max(0, j0 - l0)
    jmax = min(ny, j0 + l0)
   

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dSingle(1, n, 1, 1, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
  

    do i = imin, imax

        dx = real(i + 1) - source%xpos

        xpos = psfXcentre + dx
  
        if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
        do j = jmin, jmax
          
          
 
            dy = real(j + 1) - source%ypos
     
            ypos = psfYcentre + dy 
  
            if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
            n = n + 1
           
            xData(n, 1) = PsfInterpolation(xPos, yPos)


            if (xData(n, 1) .lt. 0.0) then
                n = max(0, n - 1)
                cycle
            end if

            ydata(n) =  image(i, j)

        
     
!            write(*, *) "i=", i, " j=", j, " dx=", dx, " dy=", dy, " xpos=", xpos, " ypos=", ypos, &
!              " psfimage=", xdata(n), " image=", ydata(n)




!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************

            weight(n) = weightingFunction( ydata(n)) 

           

        end do
    end do





    !**********************************************************************************
    ! Least-squares fit with constant (background) term
    !**********************************************************************************  
    call LeastSquaresFit2(n, xdata=xdata, ydata=ydata, &
            weight=weight, nterms=1_int8,              &
            gradient=gradient, gradienterror=gradientError, &
            constant=constant, constanterror=constantError)







!write(*, *) "PSFphotometry2 gradient=", gradient, " constant=", constant
    gradient = gradient * sumPsfImage 

!  write(*, *) "Original raw counts =", source%rawCountRate * source%exposure, & 
!" ratio (new/old)=",  (gradient / (source%rawCountRate * source%exposure) )
 

    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)




    !*******************************************************************
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !*******************************************************************
    source%rawCountRate = gradient(1) / source%exposure

    source%apertureRadius = myApertureRadius
    source%rawBackgroundCountRate = constant * PI * real(144  / &
                  (binX * binX)) / source%exposure



    call PointSourceCoincLossCorr( source, myApertureRadius, frameTime )     



    
end subroutine PSFPhotometry2




!*************************************************************************
!
! Least-squares to the PSF to include the background and an arbitrary
! number of linear constants
! ie y=constant+a1*x+a2*x, etc
!
!************************************************************************

subroutine PSFPhotometry3(nterms, image, source)
    integer(int8), intent(in) :: nTerms
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters) :: source(1 : nterms)
   

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, k, n
   
  
    real(single) :: xpos, ypos, dx, dy
    
  
   
 

    real(single), dimension(1 : nterms) :: gradient, gradientError

    real(single) :: constant, constantError


    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)



    do i = 1, nTerms

        i0 = max(0, int(source(i)%xpos - 0.5))
        j0 = max(0, int(source(i)%ypos - 0.5))

        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(nx, i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(ny, j0 + l0))
    end do

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    




    xdata => allocate2dSingle(1, n, 1, int(nterms, int32), "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************

           
            n = n + 1
            do k = 1, nTerms 
                dx = real(i + 1) - source(k)%xpos

                xpos = psfXcentre + dx
                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - source(k)%ypos
     
                ypos = psfYcentre + dy
 
             
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
             
           
           

 
                xData(n, k) = PsfInterpolation(xPos, yPos)


            end do
            if(all(xdata(n, 1: nterms) < 1e-10)) then
                n = n - 1
            else


                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(n) = image(i, j)
                



!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


                weight(n) = weightingFunction( ydata(n)) 
 
                        
                
            end if



        end do
    end do






    !**********************************************************************************
    ! Least-squares fit with constant (background) term
    !**********************************************************************************  
    call LeastSquaresFit2(n, xdata=xdata, ydata=ydata,             &
            weight=weight, nterms=int(nterms, int8),               &
            maxiter=1_int8, nstd1=3.0,                             &
            gradient=gradient, gradienterror=gradientError,        &
            constant=constant, constanterror=constantError)




   
 
    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)



    
    !*******************************************************************
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !*******************************************************************
    do i = 1, nterms    
        source(i)%rawCountRate = gradient(i) * sumPsfImage / source(i)%exposure

        source(i)%apertureRadius = myApertureRadius
        source(i)%rawBackgroundCountRate = constant * PI * real(144  / &
                  (binX * binX)) / source(i)%exposure

        call PointSourceCoincLossCorr( source(i), myApertureRadius, frameTime )     
    end do

    
end subroutine PSFPhotometry3










!*************************************************************************
!
! Least-squares to the PSF to exclude the background and an arbitrary
! number of linear constants
! ie y=a1*x+a2*x, etc
!
!************************************************************************

subroutine PSFPhotometry33(nterms, image, source, background)
    integer(int8), intent(in) :: nTerms
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters) :: source(1 : nterms)
    real(single), intent(in) :: background

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, k, n
     
    real(single) :: xpos, ypos, dx, dy
   
    real(single), dimension(1 : nterms) :: gradient, gradientError

    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    meanBackground = background

    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0


    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)


    do i = 1, nTerms

        i0 = max(0, int(source(i)%xpos - 0.5))
        j0 = max(0, int(source(i)%ypos - 0.5))

        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(nx, i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(ny, j0 + l0))
    end do

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    




    xdata => allocate2dSingle(1, n, 1, int(nterms, int32), "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************

           
            n = n + 1
            do k = 1, nTerms 
                dx = real(i + 1) - source(k)%xpos

                xpos = psfXcentre + dx
                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - source(k)%ypos
     
                ypos = psfYcentre + dy
 
             
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
             
           
                

                xData(n, k) = PsfInterpolation(xPos, yPos)


            end do
            if(all(xdata(n, 1: nterms) < 1e-10)) then
                n = n - 1
            else


                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(n) = max(0.0, image(i, j) - background) ! Remove the background
                



!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


                weight(n) = weightingFunction( ydata(n)) 
           
                
            end if



        end do
    end do






    !**********************************************************************************
    ! Least-squares fit without constant (background) term
    !**********************************************************************************  
    call LeastSquaresFit3(n, xdata=xdata, ydata=ydata,             &
            weight=weight, nterms=int(nterms, int8),               &
            maxiter=1_int8, nstd1=3.0,                             &
            gradient=gradient, gradienterror=gradientError)




   
 
    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)



    
    !*******************************************************************
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !*******************************************************************
    do i = 1, nterms    
        source(i)%rawCountRate = gradient(i) * sumPsfImage / source(i)%exposure

        source(i)%apertureRadius = myApertureRadius
        source(i)%rawBackgroundCountRate = meanBackground * PI * real(144  / &
                  (binX * binX)) / source(i)%exposure

        call PointSourceCoincLossCorr( source(i), myApertureRadius, frameTime )     
    end do

    
end subroutine PSFPhotometry33







!************************************************************************************
!
! Least-squares fit to PSF, including background, and including x and y
! for 1 or more sources
!
!***********************************************************************************
subroutine PSFPhotometry4(image, source, nterms)
    real(single), dimension(:, :), pointer :: image

    integer(int32), intent(in) :: nterms

    type(SourceParameters), intent(inout) :: source(1 : nterms)
  

    integer(int32) :: i0, j0
    integer(int32) :: n
  

 
  
   
    real(double) :: x(1 : nterms * 2)
    
!    logical(boolean) :: toFit(2)
    integer(int32) :: status 
    
    real(double) :: funcValue
    real(double), dimension(1 : nterms * 2) :: step, var
    real(double) :: simp, stopcr
    integer :: nop, maxFn, iquad, nloop

    integer :: i, j

    numberToFit = nterms


  


    nx = ubound(image, 1)
    ny = ubound(image, 2)






    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)


   
    do i = 1, nTerms
        j = (i - 1) * 2 + 1


        x(j) = source(i)%xpos 
        bl(j) = source(i)%xpos - real(maxShift / binx) 
        bu(j) = source(i)%xpos + real(maxShift / binx)

        x(j + 1) =  source(i)%ypos 
        bl(j + 1) = source(i)%ypos - real(maxShift / binx) 
        bu(j + 1) = source(i)%ypos + real(maxShift / binx)

  

        i0 = max(0, int(bl(j) - 0.5))
        j0 = max(0, int(bl(j + 1) - 0.5))
        
        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(int(nx, int32), i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(int(ny, int32), j0 + l0))
    end do


    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dSingle(1, n, 1, nterms, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")

 
!    toFit = .true.


  


!    call setNConstant( 100 )



    step = 0.02_double / dble(binx)





    maxFn = nterms * 5000
    stopCr = 0.0000001_double

    iquad = 1
    nop = nterms * 2
    nloop = 2 * nop

!****************************************************
! FOR TESTING PURPOSES
!****************************************************

!    simp = PSFFunction12(x) 



    simp = PSFFunction10(x) / 10.0




    do i = 1, 2 
        call omminim(x, step, nop, funcValue, maxFn, stopcr, nloop, iquad, &
                    simp, var, psfFunction10, status)  
        if(status .eq. 0) exit
    end do  

    if(status .ne. 0) then
        write(errstr, '(a, i5, a, e14.8)') "Status=", status, &
            " The fit may not be very good! - standard deviation of fit=", funcValue
        call message(AppMsg, Verbose)

    end if






    ! Write out the source list to the oswDataSet
    write(errstr,'(a, f7.2, a, f7.2, a, f7.2, a, f7.2 )') &
                "Old coordinates- x=", source(1)%xpos, " y=", source(1)%ypos

    call message(AppMsg, Verbose)
    
 
    write(errstr,'(a, f7.2, a, f7.2)') &
                "New coordinates- x=", x(1), " y=", x(2)
    
    call message(AppMsg, Verbose)



    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


    source(1)%apertureRadius = myApertureRadius
 
    source(1)%xpos = x(1)
    source(1)%ypos = x(2)









    call PSFPhotometry3(numberToFit, image, source)

    

end subroutine PSFPhotometry4



!************************************************************************************
!
! Least-squares fit to PSF, excluding background, but including x and y,
! for 1 or more sources
!
!***********************************************************************************
subroutine PSFPhotometry44(image, source, nterms, background)
    real(single), dimension(:, :), pointer :: image

    integer(int32), intent(in) :: nterms

    type(SourceParameters), intent(inout) :: source(1 : nterms)
    real(single), intent(in) :: background

    integer(int32) :: i0, j0
    integer(int32) :: n
  

 
  
   
    real(double) :: x(1 : nterms * 2)
    
!    logical(boolean) :: toFit(2)
    integer(int32) :: status 
    
    real(double) :: funcValue
    real(double), dimension(1 : nterms * 2) :: step, var
    real(double) :: simp, stopcr
    integer :: nop, maxFn, iquad, nloop

    integer :: i, j

    numberToFit = nterms

    meanBackground = background

  


    nx = ubound(image, 1)
    ny = ubound(image, 2)






    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)


   
    do i = 1, nTerms
        j = (i - 1) * 2 + 1


        x(j) = source(i)%xpos 
        bl(j) = source(i)%xpos - real(maxShift / binx) 
        bu(j) = source(i)%xpos + real(maxShift / binx)

        x(j + 1) =  source(i)%ypos 
        bl(j + 1) = source(i)%ypos - real(maxShift / binx) 
        bu(j + 1) = source(i)%ypos + real(maxShift / binx)

  

        i0 = max(0, int(bl(j) - 0.5))
        j0 = max(0, int(bl(j + 1) - 0.5))
        
        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(int(nx, int32), i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(int(ny, int32), j0 + l0))
    end do


    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dSingle(1, n, 1, nterms, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")

 


    write(errstr, '(a, f7.2)') "Calling non-linear least-squares fitting routine- background fixed at ", background
    call message(AppMsg, Verbose)




    step = 0.2_double / dble(binx)
    maxFn = nterms * 5000
    stopCr = 0.0000001_double

    iquad = 1
    nop = nterms * 2
    nloop = 2 * nop





    simp = PSFFunction10(x) / 10.0




    do i = 1, 2
        call omminim(x, step, nop, funcValue, maxFn, stopcr, nloop, iquad, &
                    simp, var, psfFunction11, status)
        if(status .eq. 0) exit  
    end do  

    if(status .ne. 0) then
        write(errstr, '(a, i5, a, e14.8)') "Status=", status, &
         " The fit may not be too good! - Standard deviation of fit=", funcValue
        call message(AppMsg, Verbose)
    end if







    ! Write out the source list to the oswDataSet
    write(errstr,'(a, f7.2, a, f7.2, a, f7.2, a, f7.2 )') &
                "Old coordinates- x=", source(1)%xpos, " y=", source(1)%ypos

    call message(AppMsg, Verbose)
    
 
    write(errstr,'(a, f7.2, a, f7.2)') &
                "New coordinates- x=", x(1), " y=", x(2)
    
    call message(AppMsg, Verbose)



    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


    source(1)%apertureRadius = myApertureRadius
 
    source(1)%xpos = x(1)
    source(1)%ypos = x(2)









    call PSFPhotometry33(numberToFit, image, source, background)

    

end subroutine PSFPhotometry44






!***********************************************************
! 
! Fitting subroutine used by nonlinls
! 
!**********************************************************
function PSFFunction10(x) result(functionValue)
    real(double), intent(inout) :: x(:)
    real(double) :: functionValue
   

    real(single) :: xpos, ypos, dx, dy

    real(single), dimension(1 : numberToFit) :: gradient, gradientError
    real(single) :: constant, constantError
    integer :: i, j, k, l, ndata
    real(double) :: fit, residual

    gradientError = 0.0

    constantError = 0.0

    !****************************************************
    ! Check to make sure x values are within bounds
    !***************************************************
    do i = 1, numberToFit

        if(x(i) .lt. bl(i)) then
            functionValue = 1e20
            return 

        end if
 
        if(x(i) .gt. bu(i)) then
            functionValue = 1e20
            return 
        
        end if
    end do  

    nData = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************
           
            nData = nData + 1

            do k = 1, numberToFit
                l = (k  - 1) * 2 + 1
                dx = real(i + 1) - x(l)

                xpos = psfXcentre + dx

                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - x(l + 1)
     
                ypos = psfYcentre + dy 
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
                      

                xData(nData, k) = PsfInterpolation(xPos, yPos)


            end do

            if(all(xdata(nData, 1: numberToFit) < 1e-10)) then
                nData = nData - 1

            else

                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(nData) = myimage(i, j)
                



!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************



                weight(nData) = weightingFunction( ydata(nData)) 
 
                                               
            end if

        end do
    end do



    !********************************************************************************
    ! Do the least-squares fit
    !*******************************************************************************  
    call LeastSquaresFit2(ndata, xdata(:,  1 : numberToFit), ydata=ydata,          &
            weight=weight, nterms=numberToFit, maxIter=1_int8, gradient=gradient,  &
            gradientError=gradientError, constant=constant,                        &
            constantError=constantError)



    !****************************************************************
    ! Compute the weighted residuals
    !***************************************************************
    
    functionValue = 0.0_double

    do i = 1, ndata

        fit = dble(constant)

        do k = 1, numberToFit ! Sum for each source
            fit = fit + dble(gradient(k)) * dble(xdata(i, k))
        end do
  
        residual = fit - dble(ydata(i))

        functionValue = functionValue + residual * residual * dble(weight(i))

    end do 


   functionValue = sqrt(functionvalue /dble(ndata)) 

   

!write(*, *) "PSFFUNCTION10 x=", x, " functionvalue=", functionvalue


   
end function PSFFunction10




!***********************************************************
! 
! Fitting subroutine for use with omminim
! Least-squares fit to one or more sources, including 
! positions
! 
!**********************************************************
function PSFFunction11(x) result(functionValue)
    real(double), intent(inout) :: x(:)
    real(double) :: functionValue
   

    real(single) :: xpos, ypos, dx, dy

    real(single), dimension(1 : numberToFit) :: gradient, gradientError
  
    integer :: i, j, k, l, ndata
    real(double) :: fit, residual

    gradientError = 0.0

 

    !****************************************************
    ! Check to make sure x values are within bounds
    !***************************************************
    do i = 1, numberToFit

        if(x(i) .lt. bl(i)) then
            functionValue = 1e20
            return 

        end if
 
        if(x(i) .gt. bu(i)) then
            functionValue = 1e20
            return 
        
        end if
    end do  

    nData = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************
           
            nData = nData + 1

            do k = 1, numberToFit
                l = (k  - 1) * 2 + 1
                dx = real(i + 1) - x(l)

                xpos = psfXcentre + dx 

                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - x(l + 1)
     
                ypos = psfYcentre + dy
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
                       
    

                xData(nData, k) = PsfInterpolation(xPos, yPos)


            end do

            if(all(xdata(nData, 1: numberToFit) < 1e-10)) then
                nData = nData - 1

            else

                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(nData) = myimage(i, j) - meanBackground
                


!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


                weight(ndata) = weightingFunction( ydata(nData)) 
               

                                               
            end if

        end do
    end do



    !********************************************************************************
    ! Do the least-squares fit
    !*******************************************************************************  
    call LeastSquaresFit3(ndata, xdata(:,  1 : numberToFit), ydata=ydata,          &
            weight=weight, nterms=numberToFit, maxIter=1_int8, gradient=gradient,  &
            gradientError=gradientError)



    !****************************************************************
    ! Compute the weighted residuals
    !***************************************************************
    
    functionValue = 0.0_double

    do i = 1, ndata

        fit = 0.0_double

        do k = 1, numberToFit ! Sum for each source
            fit = fit + dble(gradient(k)) * dble(xdata(i, k))
        end do
  
        residual = fit - dble(ydata(i))

        functionValue = functionValue + residual * residual * dble(weight(i))

     

    end do 
   
    functionValue = sqrt(functionvalue /dble(ndata)) 


end function PSFFunction11










!***********************************************************
! 
! Fitting subroutine used by nonlinls
! 
!**********************************************************
function PSFFunction12(x) result(functionValue)
    real(double), intent(inout) :: x(:)
    real(double) :: functionValue
   

    real(single) :: xpos, ypos, dx, dy

    real(single), dimension(1 : numberToFit) :: gradient, gradientError
    real(single) :: constant, constantError
    integer :: i, j, k, l, ndata
    real(double) :: fit

    gradientError = 0.0

    constantError = 0.0

    nData = 0

    functionValue = 0.0_double



    constant = 10.0

    do i = 1, numberToFit

        gradient(i) = real(100 * i)

    end do
 
    do i = imin, imax

        do j = jmin, jmax


            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************
           
            nData = nData + 1

            do k = 1, numberToFit
                l = (k  - 1) * 2 + 1
                dx = real(i + 1) - x(l)

                xpos = psfXcentre + dx

                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - x(l + 1)
     
                ypos = psfYcentre + dy 
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
                      

 
                xData(nData, k) = PsfInterpolation(xPos, yPos)


            end do

            if(all(xdata(nData, 1: numberToFit) < 1e-10)) then
                nData = nData - 1
            else


                fit = dble(constant)

               do k = 1, numberToFit ! Sum for each source
                   fit = fit + dble(gradient(k)) * dble(xdata(ndata, k))
               end do
  
 





                !****************************************************
                ! Data point okay
                !*************************************************** 
                myimage(i, j) = fit
                
                         
            
                                
            end if



        end do
    end do

end function PSFFunction12

!**************************************************
! Gets the 2-d image from the fits file
!**************************************************
subroutine createImageArray(dataSet,image)
    type(dataSetT), intent(in) :: dataSet
    real(single),   pointer    :: image(:, :)

    type(ArrayT)           :: handle
    integer(int8), pointer :: int8ImageP(:, :)     ! Integer8 array pointer.
    integer(int32),pointer :: int32ImageP(:, :)    ! Integer32 array pointer.
    integer(int16),pointer :: int16ImageP(:, :)    ! Integer16 array pointer.
    real(single),  pointer :: real32ImageP(:, :)   ! Real32 array pointer.
    real(double),  pointer :: real64ImageP(:, :)   ! Real64 array pointer.
    integer                :: sz(2)
    integer                :: status

    nullify(image)
    handle = array(dataSet, "PRIMARY", read)

    sz = dimensions(handle)


    if (all(sz > 0)) then
        allocate( image(0 : sz(1) - 1, 0 : sz(2) - 1), stat = status)


        select case(arrayDataType(handle))
        case(integer8)
            int8ImageP => int8Array2Data(handle)
            image = real(int8ImageP)

        case(integer16)
            int16ImageP => int16Array2Data(handle)
            image = real(int16ImageP)

        case(integer32)
            int32ImageP => int32Array2Data(handle)
            image = real(int32ImageP)

        case(real32)
            real32ImageP => real32Array2Data(handle)
            image = real32ImageP

        case(real64)
            real64ImageP => real64Array2Data(handle)
            image = real64ImageP
        end select
    end if
 
    call release(handle)

end subroutine createImageArray
!*********************************************************
!
!*********************************************************



!subroutine createQualityArray(dataSet,quality)
!    type(dataSetT), intent(in) :: dataSet
!    logical(boolean),  pointer    :: quality(:, :)

!    type(ArrayT)                :: handle
!    integer(int16), pointer :: arrayP(:, :)
!    integer(int32)              :: sz(2)
!    integer                     :: status

!    nullify(quality)
!    handle = array(dataSet, "QUALITY", read)
    




!    sz = dimensions(handle)
!    if (all(sz > 0)) then
!        arrayP => int16Array2Data(handle)

!        allocate( quality(0 : sz(1) - 1, 0 : sz(2) - 1), stat=status)

!        where (arrayP .ne. 0)
!            quality = .true.
!        elsewhere
!            quality = .false.
!        end where
!    end if

!    call release(handle)


!end subroutine createQualityArray
!*************************************************************
!
!************************************************************
real(single) function PsfInterpolation(xpos, ypos)
    real(single), intent(in) :: xPos, yPos




!    psfInterpolation = max(0.0, bilinearInterpolate(mypsfimage, xPos, yPos))
    psfInterpolation = max(0.0, gridInterpolate(xpos, ypos, psfgridfit))

end function psfInterpolation






!*********************************************************
!
!*********************************************************


real(single) function weightingFunction( data )
    real(single), intent(in) :: data
 

    if ( data .gt. 0.0 .and. intParameter("weighting") .eq. 1) then
        
        weightingFunction = 1.0 / data
    else
        weightIngFunction  = 1.0
    end if

end function weightingFunction
  
end module psfphotometry_mod










